@@ -29,6 +29,11 @@
 
 static connman_bool_t global_offlinemode = FALSE;
 
+connman_bool_t __connman_manager_get_offlinemode(void)
+{
+	return global_offlinemode;
+}
+
 static void append_profiles(DBusMessageIter *dict)
 {
 	DBusMessageIter entry, value, iter;
@@ -125,6 +130,78 @@ static void append_connections(DBusMessageIter *dict)
 	dbus_message_iter_close_container(dict, &entry);
 }
 
+static void append_available_technologies(DBusMessageIter *dict)
+{
+	DBusMessageIter entry, value, iter;
+	const char *key = "AvailableTechnologies";
+
+	dbus_message_iter_open_container(dict, DBUS_TYPE_DICT_ENTRY,
+								NULL, &entry);
+
+	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
+
+	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
+			DBUS_TYPE_ARRAY_AS_STRING DBUS_TYPE_STRING_AS_STRING,
+								&value);
+
+	dbus_message_iter_open_container(&value, DBUS_TYPE_ARRAY,
+					DBUS_TYPE_STRING_AS_STRING, &iter);
+	__connman_notifier_list_registered(&iter);
+	dbus_message_iter_close_container(&value, &iter);
+
+	dbus_message_iter_close_container(&entry, &value);
+
+	dbus_message_iter_close_container(dict, &entry);
+}
+
+static void append_enabled_technologies(DBusMessageIter *dict)
+{
+	DBusMessageIter entry, value, iter;
+	const char *key = "EnabledTechnologies";
+
+	dbus_message_iter_open_container(dict, DBUS_TYPE_DICT_ENTRY,
+								NULL, &entry);
+
+	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
+
+	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
+			DBUS_TYPE_ARRAY_AS_STRING DBUS_TYPE_STRING_AS_STRING,
+								&value);
+
+	dbus_message_iter_open_container(&value, DBUS_TYPE_ARRAY,
+					DBUS_TYPE_STRING_AS_STRING, &iter);
+	__connman_notifier_list_enabled(&iter);
+	dbus_message_iter_close_container(&value, &iter);
+
+	dbus_message_iter_close_container(&entry, &value);
+
+	dbus_message_iter_close_container(dict, &entry);
+}
+
+static void append_connected_technologies(DBusMessageIter *dict)
+{
+	DBusMessageIter entry, value, iter;
+	const char *key = "ConnectedTechnologies";
+
+	dbus_message_iter_open_container(dict, DBUS_TYPE_DICT_ENTRY,
+								NULL, &entry);
+
+	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
+
+	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
+			DBUS_TYPE_ARRAY_AS_STRING DBUS_TYPE_STRING_AS_STRING,
+								&value);
+
+	dbus_message_iter_open_container(&value, DBUS_TYPE_ARRAY,
+					DBUS_TYPE_STRING_AS_STRING, &iter);
+	__connman_notifier_list_connected(&iter);
+	dbus_message_iter_close_container(&value, &iter);
+
+	dbus_message_iter_close_container(&entry, &value);
+
+	dbus_message_iter_close_container(dict, &entry);
+}
+
 static DBusMessage *get_properties(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
@@ -171,6 +248,15 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	connman_dbus_dict_append_variant(&dict, "OfflineMode",
 				DBUS_TYPE_BOOLEAN, &global_offlinemode);
 
+	append_available_technologies(&dict);
+	append_enabled_technologies(&dict);
+	append_connected_technologies(&dict);
+
+	str = __connman_service_default();
+	if (str != NULL)
+		connman_dbus_dict_append_variant(&dict, "DefaultTechnology",
+						DBUS_TYPE_STRING, &str);
+
 	dbus_message_iter_close_container(&array, &dict);
 
 	return reply;
@@ -205,6 +291,8 @@ static DBusMessage *set_property(DBusConnection *conn,
 
 		global_offlinemode = offlinemode;
 
+		__connman_storage_save_global();
+
 		__connman_device_set_offlinemode(offlinemode);
 	} else if (g_str_equal(name, "ActiveProfile") == TRUE) {
 		const char *str;
@@ -218,6 +306,26 @@ static DBusMessage *set_property(DBusConnection *conn,
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
+static DBusMessage *get_state(DBusConnection *conn,
+					DBusMessage *msg, void *data)
+{
+	const char *str;
+
+	DBG("conn %p", conn);
+
+	if (__connman_security_check_privilege(msg,
+					CONNMAN_SECURITY_PRIVILEGE_PUBLIC) < 0)
+		return __connman_error_permission_denied(msg);
+
+	if (__connman_element_count(NULL, CONNMAN_ELEMENT_TYPE_CONNECTION) > 0)
+		str = "online";
+	else
+		str = "offline";
+
+	return g_dbus_create_reply(msg, DBUS_TYPE_STRING, &str,
+						DBUS_TYPE_INVALID);
+}
+
 static DBusMessage *add_profile(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
@@ -244,10 +352,120 @@ static DBusMessage *remove_profile(DBusConnection *conn,
 	return __connman_error_not_supported(msg);
 }
 
+static DBusMessage *request_scan(DBusConnection *conn,
+					DBusMessage *msg, void *data)
+{
+	enum connman_device_type type;
+	const char *str;
+	int err;
+
+	DBG("conn %p", conn);
+
+	dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &str,
+							DBUS_TYPE_INVALID);
+
+	if (g_strcmp0(str, "") == 0)
+		type = CONNMAN_DEVICE_TYPE_UNKNOWN;
+	else if (g_strcmp0(str, "wifi") == 0)
+		type = CONNMAN_DEVICE_TYPE_WIFI;
+	else if (g_strcmp0(str, "wimax") == 0)
+		type = CONNMAN_DEVICE_TYPE_WIMAX;
+	else
+		return __connman_error_invalid_arguments(msg);
+
+	err = __connman_element_request_scan(type);
+	if (err < 0) {
+		if (err == -EINPROGRESS) {
+			connman_error("Invalid return code from scan");
+			err = -EINVAL;
+		}
+
+		return __connman_error_failed(msg, -err);
+	}
+
+	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
+}
+
+static DBusMessage *enable_technology(DBusConnection *conn,
+					DBusMessage *msg, void *data)
+{
+	enum connman_device_type type;
+	const char *str;
+	int err;
+
+	DBG("conn %p", conn);
+
+	dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &str,
+							DBUS_TYPE_INVALID);
+
+	if (g_strcmp0(str, "ethernet") == 0)
+		type = CONNMAN_DEVICE_TYPE_ETHERNET;
+	else if (g_strcmp0(str, "wifi") == 0)
+		type = CONNMAN_DEVICE_TYPE_WIFI;
+	else if (g_strcmp0(str, "wimax") == 0)
+		type = CONNMAN_DEVICE_TYPE_WIMAX;
+	else if (g_strcmp0(str, "bluetooth") == 0)
+		type = CONNMAN_DEVICE_TYPE_BLUETOOTH;
+	else if (g_strcmp0(str, "gps") == 0)
+		type = CONNMAN_DEVICE_TYPE_GPS;
+	else
+		return __connman_error_invalid_arguments(msg);
+
+	err = __connman_element_enable_technology(type);
+	if (err < 0) {
+		if (err == -EINPROGRESS) {
+			connman_error("Invalid return code from enable");
+			err = -EINVAL;
+		}
+
+		return __connman_error_failed(msg, -err);
+	}
+
+	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
+}
+
+static DBusMessage *disable_technology(DBusConnection *conn,
+					DBusMessage *msg, void *data)
+{
+	enum connman_device_type type;
+	const char *str;
+	int err;
+
+	DBG("conn %p", conn);
+
+	dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &str,
+							DBUS_TYPE_INVALID);
+
+	if (g_strcmp0(str, "ethernet") == 0)
+		type = CONNMAN_DEVICE_TYPE_ETHERNET;
+	else if (g_strcmp0(str, "wifi") == 0)
+		type = CONNMAN_DEVICE_TYPE_WIFI;
+	else if (g_strcmp0(str, "wimax") == 0)
+		type = CONNMAN_DEVICE_TYPE_WIMAX;
+	else if (g_strcmp0(str, "bluetooth") == 0)
+		type = CONNMAN_DEVICE_TYPE_BLUETOOTH;
+	else if (g_strcmp0(str, "gps") == 0)
+		type = CONNMAN_DEVICE_TYPE_GPS;
+	else
+		return __connman_error_invalid_arguments(msg);
+
+	err = __connman_element_disable_technology(type);
+	if (err < 0) {
+		if (err == -EINPROGRESS) {
+			connman_error("Invalid return code from disable");
+			err = -EINVAL;
+		}
+
+		return __connman_error_failed(msg, -err);
+	}
+
+	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
+}
+
 static DBusMessage *connect_service(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
-	DBusMessageIter iter, array;
+	int err;
 
 	DBG("conn %p", conn);
 
@@ -255,26 +473,17 @@ static DBusMessage *connect_service(DBusConnection *conn,
 					CONNMAN_SECURITY_PRIVILEGE_MODIFY) < 0)
 		return __connman_error_permission_denied(msg);
 
-	dbus_message_iter_init(msg, &iter);
-	dbus_message_iter_recurse(&iter, &array);
-
-	while (dbus_message_iter_get_arg_type(&array) == DBUS_TYPE_DICT_ENTRY) {
-		DBusMessageIter entry, value;
-		const char *key;
-
-		dbus_message_iter_recurse(&array, &entry);
-		dbus_message_iter_get_basic(&entry, &key);
-
-		dbus_message_iter_next(&entry);
-		dbus_message_iter_recurse(&entry, &value);
-
-		switch (dbus_message_iter_get_arg_type(&value)) {
+	err = __connman_service_create_and_connect(msg);
+	if (err < 0) {
+		if (err == -EINPROGRESS) {
+			connman_error("Invalid return code from connect");
+			err = -EINVAL;
 		}
 
-		dbus_message_iter_next(&array);
+		return __connman_error_failed(msg, -err);
 	}
 
-	return __connman_error_not_implemented(msg);
+	return NULL;
 }
 
 static DBusMessage *register_agent(DBusConnection *conn,
@@ -326,13 +535,18 @@ static DBusMessage *unregister_agent(DBusConnection *conn,
 }
 
 static GDBusMethodTable manager_methods[] = {
-	{ "GetProperties",   "",      "a{sv}", get_properties   },
-	{ "SetProperty",     "sv",    "",      set_property     },
-	{ "AddProfile",      "s",     "o",     add_profile      },
-	{ "RemoveProfile",   "o",     "",      remove_profile   },
-	{ "ConnectService",  "a{sv}", "o",     connect_service  },
-	{ "RegisterAgent",   "o",     "",      register_agent   },
-	{ "UnregisterAgent", "o",     "",      unregister_agent },
+	{ "GetProperties",     "",      "a{sv}", get_properties     },
+	{ "SetProperty",       "sv",    "",      set_property       },
+	{ "GetState",          "",      "s",     get_state          },
+	{ "AddProfile",        "s",     "o",     add_profile        },
+	{ "RemoveProfile",     "o",     "",      remove_profile     },
+	{ "RequestScan",       "s",     "",      request_scan       },
+	{ "EnableTechnology",  "s",     "",      enable_technology  },
+	{ "DisableTechnology", "s",     "",      disable_technology },
+	{ "ConnectService",    "a{sv}", "o",     connect_service,
+						G_DBUS_METHOD_FLAG_ASYNC },
+	{ "RegisterAgent",     "o",     "",      register_agent     },
+	{ "UnregisterAgent",   "o",     "",      unregister_agent   },
 	{ },
 };
 
@@ -409,6 +623,57 @@ static GDBusMethodTable nm_methods[] = {
 	{ },
 };
 
+static int manager_load(void)
+{
+	GKeyFile *keyfile;
+	GError *error = NULL;
+	connman_bool_t offlinemode;
+
+	DBG("");
+
+	keyfile = __connman_storage_open();
+	if (keyfile == NULL)
+		return -EIO;
+
+	offlinemode = g_key_file_get_boolean(keyfile, "global",
+						"OfflineMode", &error);
+	if (error == NULL) {
+		global_offlinemode = offlinemode;
+
+		__connman_device_set_offlinemode(offlinemode);
+	}
+	g_clear_error(&error);
+
+	__connman_storage_close(keyfile, FALSE);
+
+	return 0;
+}
+
+static int manager_save(void)
+{
+	GKeyFile *keyfile;
+
+	DBG("");
+
+	keyfile = __connman_storage_open();
+	if (keyfile == NULL)
+		return -EIO;
+
+	g_key_file_set_boolean(keyfile, "global",
+					"OfflineMode", global_offlinemode);
+
+	__connman_storage_close(keyfile, TRUE);
+
+	return 0;
+}
+
+static struct connman_storage manager_storage = {
+	.name		= "manager",
+	.priority	= CONNMAN_STORAGE_PRIORITY_LOW,
+	.global_load	= manager_load,
+	.global_save	= manager_save,
+};
+
 static DBusConnection *connection = NULL;
 static gboolean nm_compat = FALSE;
 
@@ -420,6 +685,9 @@ int __connman_manager_init(DBusConnection *conn, gboolean compat)
 	if (connection == NULL)
 		return -1;
 
+	if (connman_storage_register(&manager_storage) < 0)
+		connman_error("Failed to register manager storage");
+
 	g_dbus_register_interface(connection, CONNMAN_MANAGER_PATH,
 					CONNMAN_MANAGER_INTERFACE,
 					manager_methods,
@@ -439,6 +707,8 @@ void __connman_manager_cleanup(void)
 {
 	DBG("conn %p", connection);
 
+	connman_storage_unregister(&manager_storage);
+
 	if (nm_compat == TRUE) {
 		g_dbus_unregister_interface(connection, NM_PATH, NM_INTERFACE);
 	}
